####################################
#	SATdb 0.1
#	Marco Gario
###################################

# Todo: 
#  + Fix: improve quite mode
#  + Fix: trim entries before adding to the db
#  + Fix: Throw error if propertyName does not exist
#  * Add function to compute aggregate values. 
#	Avg,min,max are defined in sqlite; add stdev and corr.
#  ** Find a way to define a subset of the dataset on 
#	which to perform aggregation function and plots.
#  - Normalize db


# Configure the absolute path of the databse
# baseName = "/ram/"


import sqlite3, sys

def main():
  baseName = "./"
  storeName = "store.db3"
  satdb = SATDB(storeName,baseName)
  if len(sys.argv) < 2:
    usage()
  else:
    satdb.openDB()
    satdb.parseCmd(sys.argv[1:])
    satdb.closeDB()

class SATDB:
  """A class to manage triples of values"""
  con = None
  cur = None

  def __init__(self,storeName,baseName):
    self.storeName = storeName
    self.baseName = baseName
    self.fullName = baseName+storeName
    self.archiveBase = baseName+"archive/"

    #Set the default output format
    self.format = "simple" 
    self.quiet = False

  def openDB(self):
    if self.con == None:
      if not self.quiet: print "c Opening db"
      self.con = sqlite3.connect(self.fullName)
      self.con.isolation_level = None
      self.cur = self.con.cursor()

  def closeDB(self):
    if self.con != None:
      if not self.quiet: print "c Closing db"
      self.con.close()
      self.con = None

  def createDB(self):
    print "c Creating db %s" % self.storeName
    dropTriplesTable="""drop table triples;"""
    createTriplesTable='''create table triples (instanceName text, 
		 	    propertyName text, 
			    value text, 
			    PRIMARY KEY (instanceName, propertyName, value));'''

    # Drop existing tables
    self.cur.execute(dropTriplesTable)

    # Create existing tables
    self.cur.execute(createTriplesTable)

  def backupDB(self):
    """Creates a copy of the database with the timestamp appended.
	Copies are stored in the "archive/" directory"""

    # To perform the backup:
    #   - Close the DB
    #   - Make a compressed copy with timestamp
    #   - Re-Open the DB
    #

    from shutil import copy
    from datetime import datetime
    import gzip
  
    # Backup copy is stored in "archive/$storeName-date"
    newStoreName=self.archiveBase+self.storeName+"-"
    newStoreName=newStoreName+datetime.now().strftime("%Y%m%d.%H.%M.%S")

    self.closeDB()
  
    f_in = open(self.fullName,'rb')
    f_out = gzip.open(newStoreName,'wb')
    f_out.writelines(f_in)
    f_in.close()
    
    self.openDB()
    if not self.quiet: print "c DB Backup with name %s" % newStoreName

  def getSize(self):
    self.cur.execute("""select count(*) from triples;""")
    return self.cur.fetchall()[0][0]
    
    
  def addProperty(self,instanceName,propertyName,value):
    t = (instanceName,propertyName,value)
    self.cur.execute("insert into triples values (?,?,?)",t)

  def getAll(self,instanceName=None,format="simple"):
    if (instanceName==None):
      # Return all instances
      self.cur.execute("select * from triples;")
    else:
      self.cur.execute("select * from triples where instanceName=?;",(instanceName,))

    self.printCur()

  def interactiveShell(self):
    buffer = ""

    print "Enter your SQL commands to execute in sqlite3."
    print "Enter a blank line to exit."

    while True:
      line = raw_input()
      if line == "":
          break
      buffer += line
      if sqlite3.complete_statement(buffer):
          try:
              buffer = buffer.strip()
              self.cur.execute(buffer)

              if buffer.lstrip().upper().startswith("SELECT"):
                self.printCur()
                #print self.cur.fetchall()
          except sqlite3.Error, e:
            print "An error occurred:", e.args[0]
          buffer = ""

  def addPropertiesSlow(self,propertyFile,instanceName=None):
    pFile = file(propertyFile)
    for line in pFile:
      t=line.split()
      if t[0]=="c": continue
      if (instanceName == None) and (len(t)==3):
        tuple=(t[0],t[1],t[2])
      elif (instanceName != None) and (len(t)==2):
        tuple=(instanceName,t[0],t[1])
      else:
	print "c Error parsing '",t,"'"

      try:
        self.cur.execute("insert into triples values (?,?,?)",tuple)
        print "c added '%s' '%s' '%s'" % tuple
      except sqlite3.IntegrityError:
	print "c Duplicate entry: (%s,%s,%s) - skipped " % tuple

    print "done"

    pFile.close()

  def addProperties(self,propertyFile,instanceName=None):
    pFile = file(propertyFile)
    elements=[]
    for line in pFile:
      t=line.split()
      if t[0]=="c": continue
      if (instanceName == None) and (len(t)==3):
        tuple=(t[0],t[1],t[2])
      elif (instanceName != None) and (len(t)==2):
        tuple=(instanceName,t[0],t[1])
      elements.append(tuple)
      print "c added '%s' '%s' '%s'" % tuple

    print "c committing...",
    self.cur.executemany("insert into triples values (?,?,?)",elements)
    print "done"

    pFile.close()

  def getProperty(self,propertyName,instanceName=None):
    if instanceName==None:
      self.cur.execute("select * from triples where propertyName=?;",(propertyName,))
    else:
      t=(instanceName,propertyName)
      self.cur.execute("select * from triples where instanceName=? and propertyName=?;",t)

    self.printCur()

  def delAll(self,instanceName):
    cnt = self.con.execute("delete from triples where instanceName=?",(instanceName,)).rowcount
    print "c %d rows deleted" % cnt

  def delLine(self,instanceName,propertyName):
    cnt = self.con.execute("delete from triples where instanceName=? and propertyName=?",(instanceName,propertyName)).rowcount
    print "c %d rows deleted" % cnt

  def getInstanceCount(self):
    self.cur.execute("select count() from (select distinct instanceName from triples)")
    return self.cur.fetchone()[0]

  def getInstanceList(self):
    self.cur.execute("select distinct instanceName from triples;")
    self.printCur()
   
  def join(self,properties):
    cnt = len(properties)
    query = "select t0.instanceName,"

    for i in range(0,cnt):
      query = query +(" t%d.value," %i)
    
    query = query[:-1] + " from " # Get rid of ',' and add from

    for i in range(0,cnt):
      query = query + (" (select * from triples where propertyName=?) as t%d," % i)

    query = query[:-1]

    query = query +" where "

    for i in range(0,cnt-1):
      query = query + "t%d.instanceName=t%d.instanceName and " %(i,i+1)

    query = query[:-4]

    t = tuple(properties)
    self.cur.execute(query,t)

    theHeader="instanceName,"
    for p in properties:
      theHeader += "%s," %p
    theHeader = theHeader[:-1]

    self.printCur(header=theHeader)

  def joinAll(self):
    self.cur.execute("select distinct propertyName from triples;")
    pList=[]
    for i in self.cur.fetchall():
      pList.append(i[0])

    self.join(pList)

 
  def corPlot(self,p1,p2,fileName="plot.ps"):
    query = "select t1.value, t2.value "
    query += "from (select * from triples where propertyName = ?) as t1, "
    query += " (select * from triples where propertyName = ?) as t2 "
    query += " where t1.instanceName=t2.instanceName "

    t =(p1,p2)

    self.cur.execute(query,t)

    # Prepare gnuplot file
    import subprocess
    proc = subprocess.Popen(['gnuplot','-p'], 
                        shell=True,
                        stdin=subprocess.PIPE,
                        )
    proc.stdin.write('set terminal postscript eps color enhanced \n')
    proc.stdin.write('set output "%s"\n' % fileName)
    proc.stdin.write('set xlabel "%s"\n' % p1)
    proc.stdin.write('set ylabel "%s"\n' % p2)
    proc.stdin.write('set noxtics\n')
    proc.stdin.write('plot "-" using 1:2 with points title "instance"\n')
    # Add query results
    proc.stdin.write("%s %s\n" % (p1,p2))
    for l in self.cur.fetchall():
      proc.stdin.write("%s %s\n" % l)

    proc.stdin.write('quit\n')

  def getMissingProperty(self,propertyName):
    query = "select distinct instanceName from triples where instanceName "
    query += "not in (select instanceName from triples where propertyName=?)"

    self.cur.execute(query,(propertyName,))

    self.printCur()
    
  def getPropertyList(self):
    self.cur.execute("select distinct propertyName from triples;")
    self.printCur()

  def getRatio(self,p1,p2):
    query = "select t1.instanceName, t1.value, t2.value "
    query += "from (select * from triples where propertyName = ?) as t1, "
    query += " (select * from triples where propertyName = ?) as t2 "
    query += " where t1.instanceName=t2.instanceName "

    self.cur.execute(query,(p1,p2))

    name = "%sTo%s" % (p1,p2)

    for i in self.cur.fetchall():
      v1 = float(i[1])
      v2 = float(i[2])
      ratio = (v1*100)/v2
      
      print "%s %s %s" %(i[0],name,str(ratio))
    
  def parseCmd(self,args):
    if args[0]=="-csv":
      self.format="csv"
      self.parseCmd(args[1:])
    elif args[0]=="-q":
      self.quiet=True
      self.parseCmd(args[1:])
    elif args[0]=="createDB" : 
      self.createDB()
    elif args[0]=="backupDB":
      self.backupDB()
    elif args[0]=="getSize":
      print "c Records count %d" % self.getSize()
    elif args[0]=="addProperty" and len(args)==4:
      self.backupDB()
      self.addProperty(args[1],args[2],args[3])
    elif args[0]=="getAll" and len(args)==1:
      self.getAll()
    elif args[0]=="getAll" and len(args)==2:
      self.getAll(args[1])   
    elif args[0]=="addAll" and len(args)==2 and args[1]!="-s":
      self.backupDB()
      self.addProperties(args[1])
    elif args[0]=="addAll" and len(args)==3 and args[1]!="-s":
      self.backupDB()
      self.addProperties(args[2],args[1])
    elif args[0]=="addAll" and len(args)==3 and args[1]=="-s":
      self.backupDB()
      self.addPropertiesSlow(args[2])
    elif args[0]=="addAll" and len(args)==4 and args[1]=="-s":
      self.backupDB()
      self.addPropertiesSlow(args[2],args[1])
    elif args[0]=="getProperty" and len(args)==2:
      self.getProperty(args[1])
    elif args[0]=="getProperty" and len(args)==3:
      self.getProperty(args[2],args[1])
    elif args[0]=="getMissingProperty" and len(args)==2:
      self.getMissingProperty(args[1])
    elif args[0]=="delAll" and len(args)==2:
      self.backupDB()
      self.delAll(args[1])
    elif args[0]=="delLine" and len(args)==3:
      self.backupDB()
      self.delLine(args[1],args[2])
    elif args[0]=="instanceCount":
      print "c Instance count %d" % self.getInstanceCount()
    elif args[0]=="instanceList":
      self.getInstanceList()
    elif args[0]=="propertyList":
      self.getPropertyList()
    elif args[0]=="join" and len(args)==1:
      self.joinAll()
    elif args[0]=="join" and len(args)>2:
      self.join(args[1:])
    elif args[0]=="correlationPlot" and len(args)==3:
      self.corPlot(args[1],args[2])
    elif args[0]=="ratio" and len(args)==3:
      self.getRatio(args[1],args[2])
    elif args[0]=="-i":
       self.interactiveShell()
    elif args[0]=="-h":
       help()
    else:
      usage()

  def printCur(self,format=None,header=None):
    if header==None: 
      header = ""
      for i in self.cur.description:
        header += "%s " % i[0]
      header=header[:-1]

    if format==None:
      format = self.format
    if format=="simple":
      print "c %s" % header
      for t in self.cur:
        for e in t:
          print "%s" % e,
        print ""
    elif format=="csv":
      print header
      for t in self.cur:
        line=""
        for e in t:
          line+= ("%s," % e)
        line=line[:-1]
        print line       
    else:
      print "c printing format %s not available" % format


def usage():
  thisName = sys.argv[0]
  print "Usage: %s [-csv] command instance [args]" % thisName
  print ""
  print "\t%s addProperty <instance> <propertyName> <value>" % thisName
  print "\t%s getProperty [<instance>] <propertyName>" % thisName
  print "\t%s getMissingProperty <propertyName>" % thisName
  print "\t%s addAll [-s] [<instance>] <propertyFile>" % thisName
  print "\t%s getAll [<instance>]" % thisName
  print "\t%s delLine <instance> <propertyName>" % thisName
  print "\t%s delAll <instance>" % thisName
  print "\t%s instanceCount" % thisName
  print "\t%s instanceList" % thisName
  print "\t%s propertyList" % thisName
  print "\t%s join <propertyName1> <propertyName2> [...]" % thisName
  print "\t%s correlationPlot <propertyName1> <propertyName2>" % thisName
  print "\t%s ratio <propertyName1> <propertyName2>" % thisName
#  print "\t%s compute [avg|min|max|stdev] <propertyName1> [selector]" % thisName

  
  print "\t%s createDB" % thisName
  print "\t%s backupDB" % thisName
  print "\t%s getSize" % thisName  
  print "\t%s -i" % thisName
  print "\t%s -h" % thisName
  sys.exit(-1)

def help():
  thisName = sys.argv[0]
  print "Usage: %s command instance [args]" % thisName
  print ""
  print "\t%s addProperty <instance> <propertyName> <value>" % thisName
  print "\t\t Adds a single property line to the DB"

  print "\t%s getProperty [<instance>] <propertyName>" % thisName
  print "\t\t Returns the property value for the specified "
  print "\t\t instance. If no instance is provided, the value"
  print "\t\t of the property for ALL instances is returned"

  print "\t%s getMissingProperty <propertyName>" % thisName
  print "\t\t Returns a list of instances that are missing "
  print "\t\t the given property"

  print "\t%s addAll [<instance>] <propertyFile>" % thisName
  print "\t\t Adds a list of properties provided in a file. If "
  print "\t\t <instance> is defined, the input file must contain "
  print "\t\t two columns (space separated) with the property "
  print "\t\t and its value: e.g."
  print "\t\t\tis2SAT 0"
  print "\t\t otherwise the first column must provide the name"
  print "\t\t of the instance: e.g."
  print "\t\t\ttest.cnf is2SAT 0"
  print "\t\t The option -s is used to add one entry at the time "
  print "\t\t this prevents duplicates entry in the db (safe) but "
  print "\t\t significantly slows down the process for big imports "


  print "\t%s getAll [<instance>]" % thisName
  print "\t\t Returns all the properties for the given instance."
  print "\t\t If no instance is provided, the complete DB is"
  print "\t\t presented"

  print "\t%s delLine <instance> <propertyName>" % thisName
  print "\t\t Deletes the specified property for the instance"

  print "\t%s delAll <instance>" % thisName
  print "\t\t Deletes all properties for the given instance"

  print "\t%s getInstanceCount" % thisName
  print "\t\t Returns the count of instances in the DB"

  print "\t%s getInstanceList" % thisName
  print "\t\t Returns the name of all instances in the DB"

  print "\t%s createDB" % thisName
  print "\t\t Overwrite the current DB with a new empty one"

  print "\t%s backupDB" % thisName
  print "\t\t Creates a timestamped copy of the DB"

  print "\t%s getSize" % thisName
  print "\t\t Returns the number of record in the DB"

  print "\t%s -i" % thisName
  print "\t\t Starts an SQL shell"

  sys.exit(-1)

if __name__== "__main__":
  main()

